%{

/*
A component of the cT (TM) programming environment.
(c) Copyright 1989 Carnegie Mellon University.
cT is a trademark of Carnegie Mellon University.
All rights reserved.
May not be copied without the written consent
of Carnegie Mellon University.
*/

/*  yacc specification for precedence analysis of arithmetic expressions */


#include "compute.h"
#include "ecglobal.h"
#include "cglobals.h"
#include "tglobals.h"
#include "exprdefs.h"

#ifdef ctproto
extern char FAR *TUTORalloc(long size,int abort,char *label);
int  yywrap(void);
int  yyerror(char  *str);
int  n_parenchk(int  left,int  right);
int  yyparse(void);
int  n_yaddop(int  op);
int  n_addop(int  op);
int  yylex(void);
int  genadda(int  op,long  var,int  indices);
int  n_addtoken(struct  xtoken *t);
int  addad(int  ad,long  dp,long  addr);
int  genarray(int  op,long  var);
int  n_addlit(int  ad,int FAR *vp);
#ifdef IBMPROTO
int _CDECL sprintf(char *, const char *, ...);
#endif /* IBMPROTO */
#endif /* ctproto */



#ifdef macproto
extern int sprintf(char *, char *, ...);
#endif

static long nulla = -1;

static int i;	/* work variables */
static struct xtoken t;		/* operator/adytpe */
static char c;
static long lw;


/* ******************************************************************* */

yywrap() { return(TRUE); }

/* ******************************************************************* */

yyerror(str) 
char *str;

{	
	if (exa->errnum!= 0) return(0); /* if lexical error */
	if (indexcnt !=0) {
		exa->errnum = FORMERR;
		exa->errmess = "Array has wrong number of dimensions.";
		return(0); 
	} /* indexcnt if */
	if ((parencnt == 0) && (embedcnt == 0)) {
		cI -= 1;  /* back up for caret positioning */
		exa->errnum = FORMERR;
		exa->errmess = "Bad form."; 
		if (str)
			exa->errmess = str;
	} else { 
		if (parencnt<0) {
			exa->errnum = LPARENERR;
			exa->errmess = "Missing left parenthesis."; 
		} else if (parencnt > 0) {
			exa->errnum = RPARENERR;
			exa->errmess = "Missing right parenthesis.";
			cI -= 1; 
		} else if (embedcnt < 0) {
			exa->errnum = LPARENERR;
			exa->errmess = "Missing left embed.";
		} else {
			exa->errnum = RPARENERR;
			exa->errmess = "Missing right embed.";
			cI -= 1; 
		} /* embedcnt else */
	} /* parencnt == else */
	return(0);

} /* yyerror */

/* ******************************************************************* */

n_parenchk(left,right) int left, right; {
/* return TRUE if parens are same kind (paren, bracket, brace) */
if (left == right) return(TRUE);
parencnt--;  /* force paren error */
yyerror(NULL);
return(FALSE);
}

/* ******************************************************************* */

%}

%start stat

%type <ival> cexpr
%type <ival> index

%token SKIP
%token <dval> FLITERAL 
%token <lval> ILITERAL 
%token <lval> MLITERAL
%token <lval> MLITC
%token <lval> SYSVAR         /* floating reserved word */
%token <lval> ISYSVAR		 /* integer reserved word */
%token <lval> MSYSVAR        /* marker reserved word */
%token <lval> MSYSVARA		 /* storeable marker reserved word */
%token <lval> TSYSVAR		 /* TXTYPE reserved word */
%token <defloc> GLOBALVAL    /* global variable value */
%token <defloc> GLOBALADDR   /* global variable address */
%token <defloc> LOCALVAL     /* local variable value */
%token <defloc> LOCALADDR    /* local variable address */
%token <defloc> PASSVAL      /* pass-by-address value */
%token <defloc> PASSADDR     /* pass-by-address address */
%token <defloc> GARRAYVAL     /* global array value */
%token <defloc> GARRAYADDR   /* global array address */
%token <defloc> LARRAYVAL    /* local array value */
%token <defloc> LARRAYADDR   /* local array address */
%token <defloc> PARRAYVAL    /* pass-by-value array value */
%token <defloc> PARRAYADDR   /* pass-by-value array address */
%token <defloc> GDYARRAYVAL  /* global dynamic array value */
%token <defloc> GDYARRAYADDR /* global dynamic array address */
%token <defloc> LDYARRAYVAL  /* local dynamic array value */
%token <defloc> LDYARRAYADDR /* local dynamic array address */
%token <defloc> PDYARRAYVAL  /* pass-by-addr dynamic array */
%token <defloc> PDYARRAYADDR /* pass-by-addr dynamic array */
%token <defloc> GARRAY       /* unsubscripted global array */
%token <defloc> LARRAY       /* unsubscripted local array */
%token <defloc> PARRAY       /* unsubscripted pass-by-addr array */
%token <defloc> GDYARRAY     /* unsubscripted global dynamic array */
%token <defloc> LDYARRAY     /* unsubscripted local dynamic array */
%token <defloc> PDYARRAY     /* unsubscripted pass-by-addr dynamic array */

  /* note that special LT is one greater than LT, etc. */
%token LT SPLT GT SPGT LE SPLE GE SPGE EQ SPEQ NE SPNE
%token COMMA
%token QUOTE
%token ZK /* zk or " introduces keyname */
%token <ival> KEYNAME
%token NOT MOD

%left <ival> LEFTPAREN SLEFTEMBED RIGHTPAREN SRIGHTEMBED
%right ASSIGN
%left OR 
%left AND 
%left <ival> RELOP IRELOP
%left LSHIFT RSHIFT LUNION LMASK LDIFF
%left PLUS  MINUS 
%left TIMES DIVIDE IDIVR IDIVT
%left UMINUS  UPLUS UNOT
%left EXPONENT
%left <ival> FUNCT EFUNCT
%left ARCTAN ARCTAN2 ZLENGTH ZLENGTHM ZLENGTHF ZCOMB
%left ZEXTENT ZSAMEMARK ZPRECEDE ZSEARCH ZSETMARK ZTEXTAT
%left ZRGBN ZHSVN ZHASSTYLE ZICONCODE ZICONFILE ZCLIPBOARD

%type <dval> expr stat

%%

stat	:	expr
			{ n_addop(0); }
	|	arrays
			{ n_addop(0); }
	|	SKIP
			{ n_yaddop(SKIP) ; }
	|
			{ if (exa->errnum==0) {
                       		cI--; /* for error reporting */
			exa->errnum = BLANKERR;
 			exa->errmess="Blank expression."; } };

arrays	:	GARRAYVAL
			{ genarray(GARRAY, $1); }
	|	LARRAYVAL
			{ genarray(LARRAY, $1); }
	|	PARRAYVAL
			{ genarray(PARRAY, $1); }
	|	GDYARRAYVAL
			{ genarray(GDYARRAY, $1); }
	|	LDYARRAYVAL
			{ genarray(LDYARRAY, $1); }
	|	PDYARRAYVAL
			{ genarray(PDYARRAY, $1); }
	;

zexpr	:	GLOBALVAL ASSIGN expr
			{ addad(GLOBALADDR,$1,nulla); 
			  exa->varcnt++;
			  n_yaddop(ASSIGN);
			  exa->opcnt++; exa->assigned = TRUE; }
	|	LOCALVAL ASSIGN expr
			{ addad(LOCALADDR,$1,nulla); 
			  exa->varcnt++;
			  n_yaddop(ASSIGN);
			  exa->opcnt++; exa->assigned = TRUE; }
	|	PASSVAL ASSIGN expr
			{ addad(PASSADDR,$1,nulla); 
			  exa->varcnt++;
			  n_yaddop(ASSIGN);
			  exa->opcnt++; exa->assigned = TRUE; }
	|	MSYSVAR ASSIGN expr
		{	if (($1 != ZARROWM) && ($1 != ZARROWSEL) && ($1 != ZCLIPBOARD)) {
				exa->errnum = FORMERR;
 				exa->errmess="Illegal left side of assignment.";
			} /* if */
			t.defloc = HNULL;  	/* not defined variable */
			t.constf = FALSE;	/* set not constant */
			t.type = TMARK;
			t.code= MSYSVARA;	/* store-able marker reserved word */
			t.ivalue = $1;
			t.fvalue = 0.0;
			exa->varcnt++;
			exa->opcnt++;
			exa->assigned = TRUE; 
			n_addtoken(&t);		/* add store-able reserved marker */
			n_yaddop(ASSIGN);
		}

	|	GARRAYVAL LEFTPAREN index RIGHTPAREN ASSIGN expr
			{ if (!n_parenchk($2,$4)) return(0);
			  if (!genadda(GARRAYADDR,$1,$3))
				return(0); 
			  exa->varcnt++;
			  n_yaddop(ASSIGN);
			  exa->opcnt++; exa->assigned = TRUE; }
	|	LARRAYVAL LEFTPAREN index RIGHTPAREN ASSIGN expr
			{ if (!n_parenchk($2,$4)) return(0);
			  if (!genadda(LARRAYADDR,$1,$3))
				return(0); 
			  exa->varcnt++;
			  n_yaddop(ASSIGN);
			  exa->opcnt++; exa->assigned = TRUE; }
	|	PARRAYVAL LEFTPAREN index RIGHTPAREN ASSIGN expr
			{ if (!n_parenchk($2,$4)) return(0);
			  if (!genadda(PARRAYADDR,$1,$3))
				return(0); 
			  exa->varcnt++;
			  n_yaddop(ASSIGN);
			  exa->opcnt++; exa->assigned = TRUE; }
	|	GDYARRAYVAL LEFTPAREN index RIGHTPAREN ASSIGN expr
			{ if (!n_parenchk($2,$4)) return(0);
			  if (!genadda(GDYARRAYADDR,$1,$3))
				return(0); 
			  exa->varcnt++;
			  n_yaddop(ASSIGN);
			  exa->opcnt++; exa->assigned = TRUE; }
	|	LDYARRAYVAL LEFTPAREN index RIGHTPAREN ASSIGN expr
			{ if (!n_parenchk($2,$4)) return(0);
			  if (!genadda(LDYARRAYADDR,$1,$3))
				return(0); 
			  exa->varcnt++;
			  n_yaddop(ASSIGN);
			  exa->opcnt++; exa->assigned = TRUE; }
	|	PDYARRAYVAL LEFTPAREN index RIGHTPAREN ASSIGN expr
			{ if (!n_parenchk($2,$4)) return(0);
			  if (!genadda(PDYARRAYADDR,$1,$3))
				return(0); 
			  exa->varcnt++;
			  n_yaddop(ASSIGN);
			  exa->opcnt++; exa->assigned = TRUE; }
	|	FLITERAL ASSIGN expr
			{ exa->errnum = FORMERR;
 			  exa->errmess="Can't assign to constant.";
			  exa->varcnt++;
			  exa->opcnt++; exa->assigned = TRUE; }
	|	zexpr PLUS zexpr
			{ n_yaddop(PLUS); 
			  exa->opcnt++; }
	|	zexpr MINUS zexpr
			{ n_yaddop(MINUS); 
			  exa->opcnt++; }
	|	zexpr TIMES zexpr
			{ n_yaddop(TIMES); 
			  exa->opcnt++; }
	|	zexpr DIVIDE zexpr
			{ n_yaddop(DIVIDE); 
			  exa->opcnt++; }
	|	zexpr IDIVR zexpr
			{ n_yaddop(IDIVR); 
			  exa->opcnt++; }
	|	zexpr IDIVT zexpr
			{ n_yaddop(IDIVT); 
			  exa->opcnt++; }
	|	zexpr LSHIFT zexpr
			{ n_yaddop(LSHIFT); 
			  exa->opcnt++; }
	|	zexpr RSHIFT zexpr
			{ n_yaddop(RSHIFT); 
			  exa->opcnt++; }
	|	zexpr LMASK zexpr
			{ n_yaddop(LMASK); 
			  exa->opcnt++; }
	|	zexpr LUNION zexpr
			{ n_yaddop(LUNION); 
			  exa->opcnt++; }
	|	zexpr LDIFF zexpr
			{ n_yaddop(LDIFF); 
			  exa->opcnt++; }
	|	MINUS zexpr  %prec UMINUS
			{ n_yaddop(UMINUS); exa->canstore = FALSE; }
	|	PLUS zexpr  %prec UPLUS
			{ exa->canstore = FALSE; }
	|	zexpr EXPONENT zexpr
			{ n_yaddop(EXPONENT); 
			  exa->opcnt++; }
	|	UNOT zexpr
			{ n_yaddop(NOT); }
	|	NOT LEFTPAREN expr RIGHTPAREN
			{ if (!n_parenchk($2,$4)) return(0);
                       		  n_yaddop(NOT); }
	|	ZLENGTH LEFTPAREN arrays RIGHTPAREN
			{  if (!n_parenchk($2,$4)) return(0);
                       		  n_yaddop(ZLENGTH); 
				}
	|	ZLENGTH LEFTPAREN expr RIGHTPAREN
			{	if (!n_parenchk($2,$4)) return(0);
                       		n_yaddop(ZLENGTHM);  }
	|	ARCTAN LEFTPAREN expr RIGHTPAREN
			{ if (!n_parenchk($2,$4)) return(0);
                       		  n_yaddop(ARCTAN); }
	|	ARCTAN LEFTPAREN expr COMMA expr RIGHTPAREN
			{ if (!n_parenchk($2,$6)) return(0);
                       		  n_yaddop(ARCTAN2); }
	|	ZCOMB LEFTPAREN expr COMMA expr RIGHTPAREN
		{ 	if (!n_parenchk($2,$6)) return(0);
                       	n_yaddop(ZCOMB); 
		}
	|	ZEXTENT LEFTPAREN expr COMMA expr RIGHTPAREN
		{ 	if (!n_parenchk($2,$6)) return(0);
                       	n_addop(ZEXTENT); 
			if (!exa->mfunctst) exa->canstore = FALSE;
		}
	|	ZSAMEMARK LEFTPAREN expr COMMA expr RIGHTPAREN
		{ 	if (!n_parenchk($2,$6)) return(0);
                       	n_yaddop(ZSAMEMARK); 
			if (!exa->mfunctst) exa->canstore = FALSE;
		}
	|	ZHASSTYLE LEFTPAREN expr COMMA expr RIGHTPAREN
		{ 	if (!n_parenchk($2,$6)) return(0);
                       	n_yaddop(ZHASSTYLE); 
			exa->canstore = FALSE;
		}
	|	ZICONCODE LEFTPAREN expr COMMA expr RIGHTPAREN
		{ 	if (!n_parenchk($2,$6)) return(0);
                       	n_yaddop(ZICONCODE); 
			exa->canstore = FALSE;
		}
	|	ZICONFILE LEFTPAREN expr COMMA expr RIGHTPAREN
		{ 	if (!n_parenchk($2,$6)) return(0);
                       	n_yaddop(ZICONFILE); 
			exa->canstore = FALSE;
		}
	|	ZPRECEDE LEFTPAREN expr COMMA expr RIGHTPAREN
		{ 	if (!n_parenchk($2,$6)) return(0);
                       	n_yaddop(ZPRECEDE); 
			if (!exa->mfunctst) exa->canstore = FALSE;
		}
	|	ZSEARCH LEFTPAREN expr COMMA expr RIGHTPAREN
		{ 	if (!n_parenchk($2,$6)) return(0);
                       	n_addop(ZSEARCH); 
			if (!exa->mfunctst) exa->canstore = FALSE;
		}
	|	ZSETMARK LEFTPAREN expr COMMA expr COMMA expr RIGHTPAREN
		{ 	if (!n_parenchk($2,$8)) return(0);
                       	n_yaddop(ZSETMARK); 
			if (!exa->mfunctst) exa->canstore = FALSE;
		}
	|	ZTEXTAT LEFTPAREN expr COMMA expr COMMA expr RIGHTPAREN
		{ 	if (!n_parenchk($2,$8)) return(0);
                       	n_yaddop(ZTEXTAT); 
			if (!exa->mfunctst) exa->canstore = FALSE;
		}
	|	FUNCT LEFTPAREN expr RIGHTPAREN
		{	if (!n_parenchk($2,$4)) return(0);
			n_addop(FUNCT); 
			i = $1; 		/* function op code */
			xtokd[ntokd].ivalue = i; 
			if ((i == ZFIRST) || (i == ZLAST) || (i == ZNEXT) ||
			   (i == ZPREVIOUS) || (i == ZBASE) || (i == ZSTART) || 
			   (i == ZNEXTEXPR) ||
			   (i == ZEND) || (i == ZNEXTWORD) || (i == ZNEXTLINE)) {
				if (!exa->mfunctst) exa->canstore = FALSE;
			} else exa->canstore = FALSE;
		}
	|	EFUNCT SLEFTEMBED expr SRIGHTEMBED
		{	n_addop(EFUNCT); 
			xtokd[ntokd].ivalue = $1; 
			exa->canstore = FALSE;
		}
	|	EFUNCT SLEFTEMBED expr COMMA expr SRIGHTEMBED
		{	n_addop(EFUNCT); 
			xtokd[ntokd].ivalue = ($1)+1; 
			exa->canstore = FALSE;
		}
	|	EFUNCT SLEFTEMBED expr COMMA expr COMMA expr SRIGHTEMBED
		{	n_addop(EFUNCT); 
			xtokd[ntokd].ivalue = ($1)+2; 
			exa->canstore = FALSE;
		}
	|	MOD LEFTPAREN expr COMMA expr RIGHTPAREN
			{ if (!n_parenchk($2,$6)) return(0);
                       		  n_yaddop(MOD); }
	|	ZRGBN LEFTPAREN expr COMMA expr COMMA expr RIGHTPAREN
		{ 	if (!n_parenchk($2,$8)) return(0);
                       	n_yaddop(ZRGBN); 
			exa->canstore = FALSE;
		}
	|	ZHSVN LEFTPAREN expr COMMA expr COMMA expr RIGHTPAREN
		{ 	if (!n_parenchk($2,$8)) return(0);
                       	n_yaddop(ZHSVN); 
			exa->canstore = FALSE;
		}
	|	GARRAYVAL LEFTPAREN index RIGHTPAREN
			{ if (!n_parenchk($2,$4)) return(0);
			  if (!genadda(GARRAYVAL,$1,$3))
				return(0); 
			  exa->canstore = TRUE;
			  exa->varcnt++; }
	|	LARRAYVAL LEFTPAREN index RIGHTPAREN
			{ if (!n_parenchk($2,$4)) return(0);
			  if (!genadda(LARRAYVAL,$1,$3))
				return(0); 
			  exa->canstore = TRUE;
			  exa->varcnt++; }
	|	PARRAYVAL LEFTPAREN index RIGHTPAREN
			{ if (!n_parenchk($2,$4)) return(0);
			  if (!genadda(PARRAYVAL,$1,$3))
				return(0); 
			  exa->canstore = TRUE;
			  exa->varcnt++; }
	|	GDYARRAYVAL LEFTPAREN index RIGHTPAREN
			{ if (!n_parenchk($2,$4)) return(0);
			  if (!genadda(GDYARRAYVAL,$1,$3))
				return(0); 
			  exa->canstore = TRUE;
			  exa->varcnt++; }
	|	LDYARRAYVAL LEFTPAREN index RIGHTPAREN
			{ if (!n_parenchk($2,$4)) return(0);
			  if (!genadda(LDYARRAYVAL,$1,$3))
				return(0); 
			  exa->canstore = TRUE;
			  exa->varcnt++; }
	|	PDYARRAYVAL LEFTPAREN index RIGHTPAREN
			{ if (!n_parenchk($2,$4)) return(0);
			  if (!genadda(PDYARRAYVAL,$1,$3))
				return(0); 
			  exa->canstore = TRUE;
			  exa->varcnt++; }
	|	SYSVAR
			{ addad(SYSVAR,-1L,$1);
			  exa->varcnt++; }
	|	ISYSVAR
			{ addad(ISYSVAR,-1L,$1);
			  exa->varcnt++; }
	|	TSYSVAR
			{ addad(TSYSVAR,-1L,$1);
			  exa->varcnt++; }
	|	MSYSVAR
			{ addad(MSYSVAR,-1L,$1);
			  exa->varcnt++; }
	|	GLOBALVAL
			{ addad(GLOBALVAL,$1,nulla); 
			  exa->varcnt++; }
	|	LOCALVAL
			{ addad(LOCALVAL,$1,nulla); 
			  exa->varcnt++; }
	|	PASSVAL
			{ addad(PASSVAL,$1,nulla); 
			  exa->varcnt++; }
	|	FLITERAL
			{ n_addlit(FLITERAL,(int FAR *) &($1)); }
	|	ILITERAL
			{ n_addlit(ILITERAL,(int FAR *) &($1)); }
	|	MLITERAL
			{ n_addlit(MLITERAL,(int FAR *) &($1)); }
	|	MLITC
			{ n_addlit(MLITC,(int FAR *) &($1)); }
	|	ZK KEYNAME
			{ exa->canstore = FALSE;
		           lw = $2;
			  n_addlit(ILITERAL,(int FAR *) &lw); }
	|	QUOTE KEYNAME
			{ exa->canstore = FALSE;
		           lw = $2;
			  n_addlit(ILITERAL,(int FAR *) &lw); }
	|	QUOTE MLITERAL
			{ exa->canstore = FALSE;
		           lw = $2;
			  n_addlit(MLITERAL,(int FAR *) &lw); }
	|	QUOTE MLITC
			{ exa->canstore = FALSE;
		           lw = $2;
			  n_addlit(MLITC,(int FAR *) &lw); }
	|	LEFTPAREN expr RIGHTPAREN
			{ if (!n_parenchk($1,$3)) return(0);
			  exa->canstore = FALSE;
			}
	|	SLEFTEMBED expr SRIGHTEMBED
			{ if (!n_parenchk($1,$3)) return(0);
			  exa->canstore = FALSE;
			}
	;

index	:	expr
			{ $$ = 1; }
	|	index COMMA expr
			{ $$ += 1; /* count indices */ }
	;	

cexpr	:	zexpr
			{ $$ = 0; }
	|	cexpr RELOP
			{ if ($1) { n_yaddop($1+1); $<ival>$ = ntokd; ibranf = TRUE; }
			  else $<ival>$ = -1; }
		zexpr
			{ if ($<ival>3 >= 0) { 
                                                xtokd[$<ival>3].nextr = ntokd+1;
                                             } 
			  exa->opcnt++;
			  $$ = $2; }
	;

aexpr	:	cexpr
			{ if ($1) n_yaddop($1); }
	|	aexpr AND
			{ $<ival>$ = ntokd; }
		cexpr
			{ if ($4) n_yaddop($4);
			  n_yaddop(AND);
			  xtokd[$<ival>3].nextb = ntokd;
                          ibranf = TRUE;
			  exa->opcnt++; }
	;

expr	:	aexpr
			{ ; }
	|	expr OR
			{ $<ival>$ = ntokd; }
		aexpr
			{ n_yaddop(OR);
                          xtokd[$<ival>3].nextb = ntokd;
                          ibranf = TRUE;
			  exa->opcnt++; }
	;
%%
